@@ -47,7 +47,8 @@
 #define LDAP_DEPRECATED 1
 #include <ldap.h>
 #include "nfsidmap.h"
-#include "nfsidmap_internal.h"
+#include "nfsidmap_plugin.h"
+#include "nfsidmap_private.h"
 #include "conffile.h"
 
 /* attribute/objectclass default mappings */
@@ -1100,6 +1101,9 @@ umichldap_init(void)
 	char missing_msg[128] = "";
 	char *server_in, *canon_name;
 
+	if (nfsidmap_conf_path)
+		conf_init_file(nfsidmap_conf_path);
+
 	server_in = conf_get_str(LDAP_SECTION, "LDAP_server");
 	ldap_info.base = conf_get_str(LDAP_SECTION, "LDAP_base");
 	ldap_info.people_tree = conf_get_str(LDAP_SECTION, "LDAP_people_base");
